import queue

class TrieNode :
    def __init__(self, simbols): 
        self.children = {}
        for simbol in simbols:
            self.children[simbol]=None
  
        # is_end_of_word is True if node represent the end of the word 
        self.is_end_of_word = False

class Trie:
    """
    Data structure Trie.
    See more in :
    https://en.wikipedia.org/wiki/Trie#:~:text=In%20computer%20science%2C%20a%20trie,the%20keys%20are%20usually%20strings.
    """

    def __init__(self, simbols, **kwargs):
        """"Inits the Trie
        
        If given optional parameter words, then it
        creates a Trie with given words inserted.
        """
        self.simbols = simbols
        self.root = self.getNode() 

        if  'words' in kwargs :
            for word in kwargs['words'] :
                self.insert(word)
  
    def getNode(self): 
        """Returns new trie node (initialized to NULLs) """
        return TrieNode(self.simbols)


    def insert(self, word): 
        """
        If not present, inserts key into trie 
        If the key is prefix of trie node,  
        just marks leaf node 
        """
        current_node = self.root 
        for current_simbol in word: 
            # if current character is not present 
            if not current_node.children[current_simbol]: 
                current_node.children[current_simbol] = self.getNode() 
            current_node = current_node.children[current_simbol] 
  
        # mark last node as leaf 
        current_node.is_end_of_word = True

    def search(self, word): 
        """  
        Search key in the trie 
        Returns true if key presents  
        in trie, else false 
        """
        current_node = self.root 
        for simbol in word: 
            if not current_node.children[simbol]: 
                return False
            current_node = current_node.children[simbol] 
  
        return current_node != None and current_node.is_end_of_word

    def is_prefix(self):
        """
        Search for prefixes of words
        """
        q = queue.Queue()
        q.put(self.root)
        while not q.empty() :
            current_node = q.get()
            has_childs = False
            for simbol in self.simbols:
                if current_node.children[simbol]  != None:
                    has_childs=True
                    q.put(current_node.children[simbol])
            if has_childs and current_node.is_end_of_word:
                return False

        return True
    
    def complete_word(self, word):
        """Finds if there is a way to complete the word
        to make it belong to the Trie
        """
        result  =  set()

        current_node = self.root 
        for simbol in word: 
            if not current_node.children[simbol]: 
                return result
            current_node = current_node.children[simbol] 
  
        if current_node == None :
            return result
        q = queue.Queue()
        q.put( (current_node, "") )
        while not q.empty() :
            current_node, acum_str = q.get()
            for simbol in self.simbols:
                if current_node.children[simbol]  != None:
                    q.put((current_node.children[simbol], acum_str + simbol))
            if acum_str != "" and current_node.is_end_of_word:
                result.add(acum_str)

        return result


from UnCodeGL.source import Source
from UnCodeGL.data_structures.trie import Trie
from UnCodeGL.data_structures.encoder import Encoder
from UnCodeGL.data_structures.huffman_generator import Huffman_Generator
import re

class Code:
    """
    Core class for Coding and Decoding.
    It has integrated core metrics to it.

    To create a new Code you need to pass the rule
    for the code as a dictionary. 


    Attributes:
        arity:
            An integer representing the code arity
        code_alphabet:
            A set containing the code alphabet
        source_alphabet:
            A set containing the source alphabet
    """
    __code_trie = None
    __source_trie = None
    __uniquely_decodable = None

    def __init__(self, code_dictionary):
        """"
        Initiates the class using a code dictionary
        """
        self.code = validate_dictionary(code_dictionary)
        self.code_alphabet = set()
        self.source_words = set()
        self.code_words = set()
        self.source_alphabet = set()
        self.__reverse_code ={}
        coding_output =  ''
        source_input = ''
        for key, value in self.code.items():
            self.source_words.add(key)
            coding_output += value
            source_input += key
            self.code_words.add(value)
            self.__reverse_code[value] = key
        for simbol in coding_output:
            self.code_alphabet.add(simbol)
        for simbol in source_input :
            self.source_alphabet.add(simbol)
        self.arity = len(self.code_alphabet)

        if self.is_prefix() :
            self.__decoder = Encoder(self.code_alphabet, self.__reverse_code)
        self.__encoder = Encoder(self.source_alphabet, self.code)
        
    def get_arity(self):
        """Returns arity of the code
        
        Return:
            Integer value of the arity
        """
        return self.arity

    def is_prefix(self):
        """
            Uses a Trie data structure to determine
            if the code is prefix

            Returns:
                A boolean True if the code is prefix, False
                otherwise
        """
        return self.get_code_trie().is_prefix()

    @staticmethod
    def kraft_inequality_metric(words_length, arity):
        """
        Calculates Kraft Inequality for given word lengths
        
        Measures Kraft Inequality Metric defined as 

        .. math::
            \sum_{i} r^{-l_i} 

        where r means arity, and l sub i the length of 
        the ith word
        
        Also returns if an instantaneous code can be made with said words

        Typical Usage:

        .. highlight:: python
        .. code-block:: python

            >>> metric, can_be_inst = kraft_inequality_metric(words_length=[4,4,3,6],arity=2)
            >>> print(metric, can_be_inst)
            0.265625 True
            
            
        Args:
            words_length:
                An list with the length of the words
            arity:
                Arity of the code

        Returns:
            metric , can_be_instantaneous.
            Kraft Inequality metric.
            A boolean refering to whether or not a code
            with this word lenghts and arity exists.
            This is dependant on the metric being less than 1
        """
        
        if arity == None:
            raise Exception('arity can not be null')
        if type(arity) != int :
            raise Exception('arity must be an int')
        if words_length == None:
            raise Exception('words_length can not be null')
        
        metric = 0.0
        epsilon = 0.00001
        for l in words_length:
            metric += float(arity)**float(-l)
        return metric, metric + epsilon < 1

    
    def get_code_trie(self):
        """
            Returns a Trie data structure 
            from the words in the code
        """
        if self.__code_trie == None:
            self.__code_trie = Trie(self.code_alphabet)
            for word in self.code_words:
                self.__code_trie.insert(word)
        return self.__code_trie
    
    def __repeated_C_n(self):
            """True if C_n has began a cycle, False otherwise"""
            if len(self.__C_n ) <= 3:
                return False
            for i in range(1,len(self.__C_n)-1) :
                if self.__C_n[-1] == self.__C_n[i] and self.__C_n[-2] == self.__C_n[i-1] :
                    return True
            return False

    def is_uniquely_decodable(self):
        """Returns boolean specifying whether or not is uniquely decodable
        
        For the code to be uniquely decodable means that the function 
        code that takes a string from a the source and codifies it
        is inyective. 

        Returns:
            True if the code is uniquely decodable, False otherwise
        """
        if self.__uniquely_decodable == None :
            self.__C_n = [self.code_words]
            trie_0 = Trie(self.code_alphabet , words = self.code_words )
            while True:
                new_words = set()
                trie_n = Trie(self.code_alphabet , words = self.__C_n[-1])

                for word_i in self.__C_n[-1]:
                    for new_word in trie_0.complete_word(word_i) :
                        new_words.add(new_word)
                for word_i in self.__C_n[0]:
                    for new_word in trie_n.complete_word(word_i) :
                        new_words.add(new_word)
                if new_words == set() or self.__repeated_C_n() :
                    break
                self.__C_n.append(new_words)
            
            self.__C_infty = set()
            for i in range(1,len(self.__C_n)) :
                for word in self.__C_n[i] :
                    self.__C_infty.add(word)

            self.is_uniquely_decodable = len( self.__C_infty & self.__C_n[0] ) == 0
            
        return self.is_uniquely_decodable
            
    def encode(self, sentence):
        """"Encodes a given sentence
        """
        
        return self.__encoder.encode_sentence(sentence)
    
    def __encode_brute_force(self, sentence):
        """Encodes a given sentence using bruteforce
        """
        encoding = ''
        curr_idx = 0
        sentence_length = len(sentence)
        while curr_idx < sentence_length :
            advance = False
            for word in self.source_words :
                if re.match('^'+word , sentence[curr_idx:]) :
                    curr_idx += len(word)
                    encoding += self.code[word]
                    advance = True
                    break
            if not advance :
                raise Exception('No match found to encode {} '.format(sentence[curr_idx:]))
        return encoding
        
    def decode(self, sentence, get_any = True):
        """Decodes a given sentence, in case
        multiple outputs are possible it returns 
        any of them if the get_any flag is True. 
        
        Args:
            sentence:
                Sentence to be decoded. It must compel
                with the source alphabet.
            get_any:
                True by default. It gives a string when
                multiple answers are possible. When False
                it gives the answer in a set of possible answers. 
        
        Returns:
            A string with the decoded sentence
        """
        if self.is_prefix() :
            return self.__decoder.encode_sentence(sentence)
        result = self.__decode_brute_force(sentence)

        if len(result) == 0 :
            raise Exception ('No decoding possible')

        if get_any:
            return result.pop()
        return result
    
    def __decode_brute_force(self, sentence, curr_idx = 0):
        results = set()
        # Base Case
        if curr_idx >= len(sentence):
            return ['']
        
        for word in self.code_words :
            if re.match('^'+word , sentence[curr_idx:]) :
                for result in self.__decode_brute_force(
                    sentence,
                    curr_idx = curr_idx + len(word),
                    ) :
                    results.add(self.__reverse_code[word] + result)
        return results

    def get_word_length(self, source):
        """Gets the average word lenght of the code
        for a given Source

        Args:
            source:
                A source of type Source
        
        Returns:
            A float with the value of the word lenght of 
            the code for the given source.
        """
        if type (source) != Source :
            raise Exception('source must be of type source')
        word_length = 0.0
        for word, probability in source.dictionary.items() :
            if word not in source.dictionary :
                raise Exception ('{} not part of the code source '.format(word))
            
            word_length += probability * float(len(self.code[word]))
        return word_length

    @staticmethod
    def make_huffman_code(source, code_alphabet={'0','1'}):
        """Generates an optimal Huffman Code
        
        Args:
            source:
                A source of type Source
            code_alphabet:
                Iterable with the alphabet. The items should be
                of type str
        Returns:
            Optimal huffman code for the given source and
            with arity of the given code_alphabet
        """
        if type (source) != Source :
            raise Exception('source must be of type source')

        base_nodes  = [(probability, word) for word,probability in source.dictionary.items()]
        arity = len(code_alphabet)

        while len(base_nodes)  % arity != 1 :
            base_nodes.append((float(0), ''))
        huffman = Huffman_Generator(simbols = code_alphabet, base_nodes = base_nodes, arity = len(code_alphabet))

        return Code(huffman.make_huffman_code())


def validate_dictionary(code_dictionary):
    """
    Validates that a given dictionary is good to use as a code
    """
    if type(code_dictionary) != dict :
        raise Exception('code_dictionary must be of type dict')

    for key,value in code_dictionary.items():
        if type(key) != str :
            raise Exception('code_dictionary keys must be of type str')
        if type(value) != str : 
            raise Exception('code_dictionary keys must be of type str')
    return code_dictionary

        


class EncoderNode :
    def __init__(self, simbols): 
        self.children = {}
        for simbol in simbols:
            self.children[simbol]=None
  
        self.assigned_word = None

class Encoder:
    """
    Encoder based on Trie.
    See more in :
    https://en.wikipedia.org/wiki/Trie#:~:text=In%20computer%20science%2C%20a%20trie,the%20keys%20are%20usually%20strings.
    """

    def __init__(self, simbols, dictionary):
        """"Inits the encoder
        """
        self.simbols = simbols
        self.root = self.getNode() 

        for key, values  in dictionary.items() :
            self.insert(key, values)
    
    def getNode(self): 
        """Returns new trie node (initialized to NULLs) """
        return EncoderNode(self.simbols)
        
    def insert(self, word, assigned_word): 
        """
        If not present, inserts key into trie 
        If the key is prefix of trie node,  
        just marks leaf node 
        """
        current_node = self.root 
        for current_simbol in word: 
            # if current character is not present 
            if not current_node.children[current_simbol]: 
                current_node.children[current_simbol] = self.getNode() 
            current_node = current_node.children[current_simbol] 
  
        # mark last node as leaf 
        current_node.assigned_word = assigned_word
    
    def encode_sentence(self, sentence):
        """
        Returns string with the given word to search.
        None if invalid seach
        """
        current_node = self.root 
        encoding = ''

        for simbol in sentence:
            # When it finds a word it encodes it inmediatly         
            if not current_node.children[simbol]: 
                return None
            current_node = current_node.children[simbol] 
            if current_node.assigned_word != None :
                encoding += current_node.assigned_word
                current_node = self.root
                continue
  
        return encoding



class Source:
    """
    Class for Source. Contains a dictionary with the probabilities
    of each simbol it can generate

    Attributes:
        dictionary: 
            Dictionary of probabilities. The keys are the source simbols 
            and the values the probabilities

    Initialization can be done through dictionary or base_string
    but not both
        
    Can be initiated through a given dictionary
    that contains the simbols and the probabilities
    that each symbol occurs. 

    It can also be initiated with a base string that
    will count appearances and obtain probabilities.

    Either dictionary or base_string must be passed, not both

    Args:
        dictionary:
            A dictionary with the simbols and probabilities.
            The sum of the probabilities must be 1. 
        base_string:
            A string of the usual output of the source. It will
            be assumed that is a memoryless source
    """
    
    __alphabet = None

    def __init__(self, *args, **kwargs) :
        """
        Initializes the class
        """
        if 'base_string' in kwargs and (len(args) == 1 or 'dictionary' in kwargs):
            raise Exception ('dictionary and base_string are mutually exclusive parameters')
        if 'base_string' in kwargs :
            self.dictionary = make_dictionary_from_str(kwargs['base_string'])
            return
        if len(args) > 0 :
            self.dictionary = verify_dictionary(args[0])
            return
        if 'dictionary' in kwargs :
            self.dictionary = verify_dictionary(kwargs['dictionary'])
            return
        raise Exception ('No dictionary or base_string found to make the source')
    
    def get_alphabet(self):
        """
        Returns list with simbols of the source
        """
        if self.__alphabet == None :
            self.__alphabet = set()
            for simbol in self.dictionary :
                self.__alphabet.add(simbol)
        return self.__alphabet
    
    def get_simbol_prob(simbol):
        """
        Returns probability for a simbol
        """
        default = 0.0
        if simbol in self.dictionary:
            return dictionary[simbol]
        return default

    def get_entropy(self):
        """Calculates entropy of the source

        Entropy is given by 

        .. math::
                -\sum_{i} p(x_i)\log(p(x_i)) 
        
        Returns:
            Float with the value of the entropy.
        """
        total = 0.0


def make_dictionary_from_str(base_string):
    """Makes a dictionary with probabilities normalized"""
    dictionary = {}
    # Counting occurences
    for simbol in base_string:
        if simbol not in dictionary:
            dictionary[simbol] = 0
        dictionary[simbol] += 1
    # Normalizing
    for key,value in dictionary.items():
        dictionary[key] = float(value) / float(len(base_string))
    return dictionary

def verify_dictionary(dictionary):
    """Verify the sum of the probabilites is 1 to an epsilon"""
    total = 0.0
    epsilon = 0.00001
    for key,value in dictionary.items():
        total += float(value)
    if abs(total - 1.0) > epsilon :
        raise Exception ('dictionary probabilites does not add up to 1')
    return dictionary
    
        
        



from setuptools import find_packages, setup

setup(name='UnCodeGL',
      version='0.0.3',
      description='Library for Coding Theory',
      author= 'Gabriel Octavio Lozano Pinzon',
      author_email="golozanop@unal.edu.co",
      url="https://github.com/Matematikoi/UnCodeGL",
      install_requires=[],
      packages=find_packages(),
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      python_requires='>=3.6'
      )


import queue


class Huffman_Node :
    def __init__(self, simbols, probability, word=None): 
        self.children = {}
        for simbol in simbols:
            self.children[simbol]=None
        self.probability = probability
        self.assigned_word = word

class Huffman_Generator :
    """
    Utility class to generate Huffman Code
    """
    def __init__(self, simbols, base_nodes , arity):
        self.base_nodes = base_nodes
        self.simbols = [simbol for simbol in simbols]
        self.arity = arity
    
    def make_node(self , probability, word = None): 
        """Returns new node (initialized to NULLs) """
        return Huffman_Node(self.simbols, probability, word=word)
    
    def make_huffman_code(self):
        nodes = [self.make_node(prob, word = word) for prob,word in self.base_nodes]
        # Huffman Algorithm
        while len(nodes) > 1 :
            nodes.sort(key=lambda x : x.probability)
            new_node = self.make_node(0.0)
            new_prob = 0.0
            for i in range(self.arity):
                if nodes[i].probability > 0.0 :
                    new_node.children[self.simbols[i]] = nodes[i]
                new_prob += nodes[i].probability
            new_node.probability = new_prob
            nodes.append(new_node)
            nodes = nodes[self.arity:]
        # Make the dictionary with BFS
        self.root = nodes[0]
        result = {}
        q = queue.Queue()
        q.put((self.root,''))
        while not q.empty() :
            current_node , acum = q.get()
            if current_node.assigned_word :
                result[current_node.assigned_word] = acum 
            for simbol in self.simbols:
                if current_node.children[simbol]  != None:
                    q.put((current_node.children[simbol], acum + simbol))

        return result